//  AppRunSettings
//
//  This class provides the Application Execution settings for the QMove application
//
//  Message Identifier:      QMV
//  Message Range:           0001 - 0099
//  Trace area assignment:   None
//  Version:                 1.4.2  Build: 37
//
//  Version History:
//
//  1.0.0		- 14/11/2009		-	Initial Version
//  1.1.0		- 01/12/2009		-	Enhanced move operation modes supported
//  1.2.0		- 01/01/2010		-	Variable delay between transactions
//									-	Variable buffer size for copying
//  1.4.0		- 04/11/2010		-	Updated for 3.10.1 kernel & asynch transactions
//  1.4.2		- 22/10/2011		-	VS 2011 Updates
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2009 - 2011
//

//  Application Includes
#include "AppRunSettings.h"

//  Constructor
AppRunSettings::AppRunSettings(int argc, char * argv[])
{
	//  Constructor
	if (SetDefaults())											//  Set the runtime defaults
	{
		IsValid = ValidateParameters(argc, argv);				//  Validate and capture the parameters
	}
	if (IsValid)
	{
		if (LogLevel > LOGLEVEL_NORMAL)
		{
			//  Show the current parameter settings
			ShowSettings();
		}
	}
}

//  Destructor
AppRunSettings::~AppRunSettings(void)
{
}

//  SetDefaults
//
//  This function will set the default values for the Application settings
//
//  Parameters
//
//  Returns
//
//  BOOL  -				TRUE if the defaults were set
//						FALSE if the defaults were not set
//
//  API Use Notes:
//
//
BOOL AppRunSettings::SetDefaults(void)
{
	//  Set the default values for all settings
	szRepServer[0] = '\0';										//  Control Server
	szRepDb[0] = '\0';											//  Control Database
	RunningAsAddin = TRUE;										//  Running as server Addin Task
	AllowMultipleAddins = TRUE;									//  Allow multiple addins
	NoRepository = FALSE;										//  Allow use of the repository
	NoAppLog = FALSE;											//  Allow Application Event Logging
	EchoLog = FALSE;											//  Do not Echo to the console
	NeedsMQ = TRUE;												//  Processor needs an MQ
	LogLevel = LOGLEVEL_NORMAL;									//  Default logging level
	CreateRepository = FALSE;									//  Do not Create the repository
	UpdateDesign = FALSE;										//  Do not Update the design of the repository
	TraceArea = 0;												//  For Trace level logging - trace all areas
	TransactionDelay = DEFAULT_IT_DELAY;						//  Set default delay interval
	TransferBufferK = DEFAULT_BFRK;								//  Set Default Transfer Buffer size
	ThreadCount = DEFAULT_THREADCOUNT;							//  Default Thread count
	MaxCTX = DEFAULT_TXCOUNT;									//  Default Max Concurrent transactions
	UseLFC = FALSE;												//  Disable the large file copier
	ArenaSize = DEFAULT_ARENASIZE;								//  Default arena size for copying
	return TRUE;
}

//  ValidateParameters
//
//  This function will validate and the application command line parameters and set the appropriate
//  settings for the program execution.
//
//  Parameters
//
//  Returns
//
//  BOOL  -				TRUE if the parameters were valid
//						FALSE if the parameters were invalid
//
//  API Use Notes:
//
//
BOOL AppRunSettings::ValidateParameters(int argc, char* argv[])
{
	BOOL		bFlagIsValid;									//  Flag is valid (or not)
	int			iIndex;											//  Index in the parameters
	char		szFlag[MAX_FLAG + 1];							//  Copy of switch parameter
	char		szLocalMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Validate and capture the runtime parameters

	AllowExecution = TRUE;

	if (argc < 2)
	{
		//  Allow request to display program usage
		if (argc == 1)
		{
			ShowUsage();
			AllowExecution = FALSE;
			return TRUE;
		}
		std::cout << MSG_QMV0001E << std::endl;
		ShowUsage();
		return FALSE;
	}

	if ((argc == 2) && (strcmp(argv[1], "-?") == 0))
	{
		ShowUsage();
		AllowExecution = FALSE;
		return TRUE;
	}

	//  Copy the positional parameters

	strcpy_s(szRepDb, MAX_DATABASE, argv[1]);
	if (szRepDb[0] == '-')
	{
		std::cout << MSG_QMV0001E << std::endl;
		ShowUsage();
		return FALSE;
	}

	//  Parse any switches that were supplied

	if (argc > 2)
	{
		for (iIndex = 2; iIndex < argc; iIndex++)
		{
			bFlagIsValid = FALSE;
			strcpy_s(szFlag, MAX_FLAG, argv[iIndex]);
			if (memcmp(szFlag, "-", 1) == 0)
			{
				//  Parameter is a switch
				if (memcmp(szFlag + 1, "?", 1) == 0)
				{
					//  Show Usage - do not run the application
					ShowUsage();
					AllowExecution = FALSE;
					return FALSE;
				}

				if ((memcmp(szFlag + 1, "I", 1) == 0) || (memcmp(szFlag + 1, "i", 1) == 0))
				{
					//  Transaction Delay time in seconds
					if (memcmp(szFlag + 2, ":", 1) == 0)
					{
						TransactionDelay = atoi(szFlag + 3);
						if ((TransactionDelay < MIN_DELAY) || (TransactionDelay > MAX_DELAY))
						{
							sprintf_s(szLocalMsg, MAX_MSG, MSG_QMV0004E, TransactionDelay);
							std::cout << szLocalMsg << std::endl;
							return FALSE;
						}
						bFlagIsValid = TRUE;
					}
				}

				if ((memcmp(szFlag + 1, "B", 1) == 0) || (memcmp(szFlag + 1, "b", 1) == 0))
				{
					//  Transfer Buffer size in Kb
					if (memcmp(szFlag + 2, ":", 1) == 0)
					{
						TransferBufferK = atoi(szFlag + 3);
						if ((TransferBufferK < MIN_BFRK) || (TransferBufferK > MAX_BFRK))
						{
							sprintf_s(szLocalMsg, MAX_MSG, MSG_QMV0008E, TransferBufferK);
							std::cout << szLocalMsg << std::endl;
							return FALSE;
						}
						bFlagIsValid = TRUE;
					}
				}

				if ((memcmp(szFlag + 1, "A", 1) == 0) || (memcmp(szFlag + 1, "a", 1) == 0))
				{
					//  Arena Size in Mb
					if (memcmp(szFlag + 2, ":", 1) == 0)
					{
						ArenaSize = atoi(szFlag + 3);
						bFlagIsValid = TRUE;
					}
				}

				if ((memcmp(szFlag + 1, "M", 1) == 0) || (memcmp(szFlag + 1, "m", 1) == 0))
				{
					//  Thread Count
					if (memcmp(szFlag + 2, ":", 1) == 0)
					{
						ThreadCount = atoi(szFlag + 3);
						if ((ThreadCount < MIN_THREADCOUNT) || (ThreadCount > MAX_THREADCOUNT))
						{
							sprintf_s(szLocalMsg, MAX_MSG, MSG_QMV0010E, ThreadCount);
							std::cout << szLocalMsg << std::endl;
							return FALSE;
						}
						bFlagIsValid = TRUE;
					}
				}

				if ((memcmp(szFlag + 1, "X", 1) == 0) || (memcmp(szFlag + 1, "x", 1) == 0))
				{
					//  Max concurrent transactions
					if (memcmp(szFlag + 2, ":", 1) == 0)
					{
						MaxCTX = atoi(szFlag + 3);
						bFlagIsValid = TRUE;
					}
				}

				if ((memcmp(szFlag + 1, "L", 1) == 0) || (memcmp(szFlag + 1, "l", 1) == 0))
				{
					//  Enable the Large File Copier
					UseLFC = TRUE;
					bFlagIsValid = TRUE;
				}

				if ((memcmp(szFlag + 1, "V", 1) == 0) || (memcmp(szFlag + 1, "v", 1) == 0))
				{
					//  Verbose logging
					LogLevel = LOGLEVEL_VERBOSE;
					bFlagIsValid = TRUE;
				}

				if ((memcmp(szFlag + 1, "D", 1) == 0) || (memcmp(szFlag + 1, "d", 1) == 0))
				{
					//  Debug logging
					LogLevel = LOGLEVEL_DEBUG;
					//  Debug Area
					if (memcmp(szFlag + 2, ":", 1) == 0)
					{
						TraceArea = atoi(szFlag + 3);
					}
					bFlagIsValid = TRUE;
				}

				if ((memcmp(szFlag + 1, "T", 1) == 0) || (memcmp(szFlag + 1, "t", 1) == 0))
				{
					//  Trace logging
					LogLevel = LOGLEVEL_TRACE;
					//  Trace Area
					if (memcmp(szFlag + 2, ":", 1) == 0)
					{
						TraceArea = atoi(szFlag + 3);
					}
					bFlagIsValid = TRUE;
				}

				if ((memcmp(szFlag + 1, "E", 1) == 0) || (memcmp(szFlag + 1, "e", 1) == 0))
				{
					//  Echo Logging to console
					EchoLog = TRUE;
					bFlagIsValid = TRUE;
				}

				if (!bFlagIsValid)
				{
					//  Invalid switch
					sprintf_s(szLocalMsg, MAX_MSG, MSG_QMV0003E, szFlag);
					std::cout << szLocalMsg << std::endl;
				}
			}
			else
			{
				//  Allow an "=" switch (override default ini file search on notes initialisation)
				if (memcmp(szFlag, "=", 1) != 0)
				{
					//  Invalid parameter
					sprintf_s(szLocalMsg, MAX_MSG, MSG_QMV0002E, szFlag);
					std::cout << szLocalMsg << std::endl;
				}
			}
		}
	}
	return TRUE;
}

//  ShowUsage
//
//  This function will show the application command line usage pattern.
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//
void AppRunSettings::ShowUsage(void)
{
	//  Show application usage

	std::cout << MSG_USAGE1 << std::endl;
	std::cout << MSG_USAGE2 << std::endl;
	std::cout << MSG_USAGE3 << std::endl;
	std::cout << MSG_USAGE4 << std::endl;
	std::cout << MSG_USAGE5 << std::endl;
	std::cout << MSG_USAGE6 << std::endl;
	std::cout << MSG_USAGE7 << std::endl;
	std::cout << MSG_USAGE8 << std::endl;
	std::cout << MSG_USAGE9 << std::endl;
	std::cout << MSG_USAGE10 << std::endl;
	std::cout << MSG_USAGE11 << std::endl;
	std::cout << MSG_USAGE12 << std::endl;
}

//  ShowSettings
//
//  This function will show the current application run settings.
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//
void AppRunSettings::ShowSettings(void)
{
	//  Show the current Run Settings

	char		szLocalMsg[MAX_MSG + 1];							//  Generic message buffer

	//  Show the repository scope
	sprintf_s(szLocalMsg, MAX_MSG, MSG_QMV0005I, szRepDb);
	std::cout << szLocalMsg << std::endl;

	//  Show the Transaction Delay
	sprintf_s(szLocalMsg, MAX_MSG, MSG_QMV0007I, TransactionDelay);
	std::cout << szLocalMsg << std::endl;

	//  Show the Transfer Buffer Size
	sprintf_s(szLocalMsg, MAX_MSG, MSG_QMV0009I, TransferBufferK);
	std::cout << szLocalMsg << std::endl;

		//  Show the Thread Count
	sprintf_s(szLocalMsg, MAX_MSG, MSG_QMV0011I, ThreadCount);
	std::cout << szLocalMsg << std::endl;

	//  Show the Max Transaction Setting
	sprintf_s(szLocalMsg, MAX_MSG, MSG_QMV0012I, MaxCTX);
	std::cout << szLocalMsg << std::endl;

	//  Show the logging level
	if (LogLevel == LOGLEVEL_NORMAL)
	{
		sprintf_s(szLocalMsg, MAX_MSG, MSG_QMV0006I, "Normal");
	}
	else
	{
		if (LogLevel == LOGLEVEL_VERBOSE)
		{
			sprintf_s(szLocalMsg, MAX_MSG, MSG_QMV0006I, "Verbose");
		}
		else
		{
			if (LogLevel == LOGLEVEL_TRACE)
			{
				sprintf_s(szLocalMsg, MAX_MSG, MSG_QMV0006I, "Trace");
			}
			else
			{
				sprintf_s(szLocalMsg, MAX_MSG, MSG_QMV0006I, "Debug");
			}
		}
	}
	std::cout << szLocalMsg << std::endl;
}
