//  AppTransactionHandler
//
//  This class provides the custom Transaction Queue handling for the QMove application
//
//  Message Identifier:      QMV
//  Message Range:           0300 - 0399
//  Trace area assignment:   None
//  Version:                 1.4.2  Build: 37
//
//  Version History:
//
//  1.4.0	-	04/11/2010		-	Transaction Handler introduced
//  1.4.1	-	21/12/2010		-	Switch to Runtime 3.10.3
// 									Fixed an issue with the transaction handler
//  1.4.2	-	22/10/2011		-	VS 2011 Updates
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2010 - 2011
//

//  Application Includes
#include "AppTransactionHandler.h"

//  Default constructor
AppTransactionHandler::AppTransactionHandler(void)
{
}

//  Normal constructor
AppTransactionHandler::AppTransactionHandler(MTExecutive *xeParent, int iThreadID)
{
	//  Save the Runtime Environment
	xeLocal = xeParent;
	ParentThread = iThreadID;

	//  Clear the throughput stats
	dwCompleted = 0;
	dwError = 0;
	dwRetried = 0;
	dwDelayed = 0;
}

//  Destructor
AppTransactionHandler::~AppTransactionHandler(void)
{
}

//  SerializeTransaction
//
//  This function will :-
//  1)  Mark the transaction document that this object refers to with a completion status
//  2)  Delete the transaction object
//
//  Parameters
//
//  1  -  void *				-  Pointer to the transaction object
//  2  -  DBHANDLE				-  Handle of the transaction queue database
//  3  -  TransactionQueue *	-  Pointer to the Transaction Queue object for this transaction
//  4  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void AppTransactionHandler::SerializeTransaction(void *txObject, HANDLE hdbQueue, TransactionQueue *tqCurrent, int iThreadID)
{
	STATUS				stAPIRC;										//  Return code from Notes API calls
	DBHANDLE			hdbLocalRep;									//  Thread Local handle for the repository
	NOTEHANDLE			hnTransaction;									//  Handle of the Transaction Note
	NOTEHANDLE			hnNewTX;										//  Handle of a new Transaction Note
	QMoveRequest		*qmrOld;										//  QMove Request
	PartMoveRequest		*pmrCarrier;									//  Address of carrier object
	TIMEDATE			tdCurrent;										//  Current Timestamp
	char				szMsg[MAX_MSG + 1];								//  Generic Message Buffer

	//  Cast the returned object to a QMove Request
	pmrCarrier = (PartMoveRequest *) txObject;
	qmrOld =  (QMoveRequest *) pmrCarrier->Parent;
	delete pmrCarrier;

	//  Open the transaction note
	stAPIRC = NSFNoteOpen(hdbQueue, qmrOld->nidRequest, 0, &hnTransaction);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_QMV0304E, qmrOld->szTransID, qmrOld->nidRequest, tqCurrent->szQName);
		xeLocal->LogMessage(szMsg, iThreadID);
		delete qmrOld;
		return;
	}

	//  Set the move finish time timestamp
	OSCurrentTIMEDATE(&tdCurrent);
	stAPIRC = NSFItemSetTime(hnTransaction, TRANSDOC_FINISHTIME, &tdCurrent);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_QMV0302E);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
	}

	if (qmrOld->ReturnCode == RETURN_NOERROR)
	{
		//  Transaction completed ok

		//  Write the stats to the request document
		qmrOld->writeStats(hnTransaction);

		//  Mark the transaction as completed
		MarkTransactionCompleted(qmrOld->nidRequest, hnTransaction, tqCurrent, iThreadID);
		dwCompleted++;
	}
	else
	{
		//  Transaction failed

		//  determine if the current transaction can be retried
		if (qmrOld->AllowRetry)
		{
			//  Check that we have not exceeded our retry limit
			if (qmrOld->RetryCount < qmrOld->RetryLimit)
			{
				//  Mark the transaction as retried
				MarkTransactionRetried(qmrOld->nidRequest, hnTransaction, tqCurrent, iThreadID);
				dwRetried++;

				//  Create a new transaction document for the retry request
				hdbLocalRep = xeLocal->GetMappedDBH(xeLocal->hdbRepository, iThreadID);
				stAPIRC = NSFNoteCreate(hdbLocalRep, &hnNewTX);
				if (ERR(stAPIRC) != NOERROR)
				{
					sprintf_s(szMsg, MAX_MSG, MSG_QMV0319E, qmrOld->szTransID);
					LogTransactionMessage(szMsg, qmrOld->hTXLog, iThreadID);
					xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
					LogTransactionMessage(szMsg, qmrOld->hTXLog, iThreadID);

					//  Remark the original request as in error
					MarkTransactionError(qmrOld->nidRequest, hnTransaction, tqCurrent, iThreadID);
					dwError++;
					dwRetried--;
				}
				else
				{
					//  Increment the retry count
					qmrOld->RetryCount++;

					//  Seralize the transaction data to the new request document
					if (!qmrOld->serialize(0, hnNewTX, xeLocal))
					{
						sprintf_s(szMsg, MAX_MSG, MSG_QMV0320E, qmrOld->szTransID);
						LogTransactionMessage(szMsg, qmrOld->hTXLog, iThreadID);

						NSFNoteClose(hnNewTX);
						//  Remark the original request as in error
						MarkTransactionError(qmrOld->nidRequest, hnTransaction, tqCurrent, iThreadID);
						dwError++;
						dwRetried--;
					}
					else
					{
						//  Get the NoteID
						NSFNoteGetInfo(hnNewTX, _NOTE_ID, &qmrOld->nidRequest);
				
						//  Set the Urgent & Approved flags
						NSFItemSetText(hnNewTX, TRANSDOC_URGENT_FLAG, "0", MAXWORD);
						NSFItemSetText(hnNewTX, TRANSDOC_APPROVED_FLAG, "Yes", MAXWORD);

						//  Mark the new transaction as delayed
						MarkTransactionDelayed(qmrOld->nidRequest, hnNewTX, FALSE, tqCurrent, iThreadID);
						dwDelayed++;

						//  Save and close the new transaction
						NSFNoteUpdate(hnNewTX, UPDATE_FORCE);
						NSFNoteClose(hnNewTX);

						sprintf_s(szMsg, MAX_MSG, MSG_QMV0321I, qmrOld->szTransID);
						xeLocal->LogMessage(szMsg, iThreadID);
					}					
				}
			}
			else
			{
				//  The transaction has exceeded the retry count and will be marked in error
				sprintf_s(szMsg, MAX_MSG, MSG_QMV0322E, qmrOld->szTransID, qmrOld->RetryLimit);
				xeLocal->LogMessage(szMsg, iThreadID);
				MarkTransactionError(qmrOld->nidRequest, hnTransaction, tqCurrent, iThreadID);
				dwError++;
			}
		}
		else
		{
			//  Retries are not allowed - mark the transaction in error
			MarkTransactionError(qmrOld->nidRequest, hnTransaction, tqCurrent, iThreadID);
			dwError++;
		}
	}

	//  Append the history log to the transaction
	AddHistory(qmrOld->hTXLog, hnTransaction, iThreadID);

	//  Dispose of the transaction object
	delete qmrOld;

	//  Save and close the transaction
	NSFNoteUpdate(hnTransaction, UPDATE_FORCE);
	NSFNoteClose(hnTransaction);

	return;
}

//  MarshallTransaction
//
//  This function will create the request object from a transaction and set the values needed to
//  dispatch it.
//
//  Parameters
//
//  1  -  NOTEID				-  Note ID of the transaction note
//  2  -  NOTEHANDLE			-  Handle of the open transaction note
//  3  -  TransactionQueue *	-  Pointer to the Transaction Queue object for this transaction
//  4  -  void **				-  Pointer for the returmed address to the transaction object
//  5  -  void **				-  Pointer for the returned address for the execution object
//  6  -  UINT *				-  Pointer for the returned Attributes for the transaction
//  7  -  int *					-  Pointer for the returned Priority for the transaction
//  8  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//  int			-   The transaction disposition
//
//	TX_DISPOSITION_NORMAL - Dispatch the transaction object and mark the transaction In Progress
//	TX_DISPOSITION_ERROR  - Mark the transaction Error
//  TX_DISPOSITION_DELAY  - Mark the transaction delayed
//  TX_DISPOSITION_PANIC  - Critical Error, stop the queue
//
//  API Use Notes:
//
//
//
int AppTransactionHandler::MarshallTransaction(NOTEID nidTX, NOTEHANDLE hnTX, TransactionQueue *tqCurrent, void **txObject, void **xxObject, UINT *Attrs, int *Priority, int iThreadID)
{
	STATUS				stAPIRC;										//  Return code fron Notes API calls
	QMoveRequest		*qmrNew;										//  QMove Request
	PartMoveRequest		*pmrCarrier;									//  Carrier PartMoveRequest
	AppRunSettings		*arsRun;										//  Settings for this run
	TIMEDATE			tdCurrent;										//  Current Timestamp
	WORD				wRetLen;										//  Returned length
	char				*pList;											//  Pointer to the text history list
	char				szMsg[MAX_MSG + 1];								//  Generic Message Buffer

	//  Remove the old transaction history log from the transaction
	stAPIRC = NSFItemDelete(hnTX, TRANSDOC_HISTORY, (WORD) strlen(TRANSDOC_HISTORY));
	if ((ERR(stAPIRC) != NOERROR) && (ERR(stAPIRC) != ERR_ITEM_NOT_FOUND))
	{
		xeLocal->LogMessage(MSG_QMV0309E, iThreadID);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
	}

	//  Set the move start time timestamp
	OSCurrentTIMEDATE(&tdCurrent);
	stAPIRC = NSFItemSetTime(hnTX, TRANSDOC_STARTTIME, &tdCurrent);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_QMV0301E);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
	}

	//  Create a New QMove Request
	qmrNew = new QMoveRequest();

	//  Create a new Text List to hold the Transaction Log History
	stAPIRC = ListAllocate(0, 0, FALSE, &qmrNew->hTXLog, &pList, &wRetLen);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_QMV0306E, iThreadID);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		qmrNew->hTXLog = NULLHANDLE;
	}

	//  UnLock the memory for the list
	OSUnlockObject(qmrNew->hTXLog);

	//  Populate the request from the transaction document
	if (!qmrNew->populate(nidTX, hnTX))
	{
		sprintf_s(szMsg, MAX_MSG, MSG_QMV0303E, nidTX);
		LogTransactionMessage(szMsg, qmrNew->hTXLog, iThreadID);
		//  Append the history log to the transaction
		AddHistory(qmrNew->hTXLog, hnTX, iThreadID);
		delete qmrNew;
		return TX_DISPOSITION_ERROR;
	}

	//  Check the availability of the source and target servers
	if (checkAvailability(qmrNew, iThreadID) != STATE_EXEC_REQUEST)
	{
		//  The transaction should be delayed
		dwDelayed++;
		delete qmrNew;
		return TX_DISPOSITION_DELAY;
	}

	//  Set up the rest of the information required by the interface
	qmrNew->PermitRun = TRUE;							//  Set the Run Permit

	//  Fill the copy buffer size
	arsRun = (AppRunSettings *) xeLocal->rsLocal;
	qmrNew->CopyBfrKb = arsRun->TransferBufferK;
	qmrNew->ArenaSize = arsRun->ArenaSize;

	//  Disable the Large File Copier - if requested
	if (!arsRun->UseLFC) qmrNew->UseLFC = FALSE;
	else qmrNew->UseLFC = TRUE;

	//  The Move Request is sent as the parent to a low level move request
	pmrCarrier = new PartMoveRequest(qmrNew);
	pmrCarrier->MoveAction = MOVE_DATABASE;
	*txObject = pmrCarrier;										//  Address of the request object

	//  Determine if this is a move transaction or a feeder transaction and dispatch accordingly
	if ((!xeLocal->IsThisADatabase(qmrNew->szSourceDatabase)) || (HasWildcards(qmrNew->szSourceDatabase)))
	{
		//  Feeder transaction
		*xxObject = qfMaster;								//  Processed by the QMFeeder engine
		*Attrs = RQATTR_REJOIN | RQATTR_CMLSL1;
		*Priority = 20;
	}
	else
	{
		//  Move transaction
		
		*xxObject = qmMaster;								//  Processed by the QCopier engine
		*Attrs = RQATTR_REJOIN | RQATTR_CMLSL2;
		*Priority = 100;
	}
	return TX_DISPOSITION_NORMAL;
}

//  LogTransactionMessage
//
//  This function will append a decorated version of the message to the Transaction History Log.
//  It will also pass the message to the system logger.
//
//  Parameters
//
//  1  -  char *				-  Pointer to the undecorated message text
//  2  -  HANDLE				-  Handle of the Transaction History Text List
//  3  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//
//  API Use Notes:
//
void AppTransactionHandler::LogTransactionMessage(char *szMsg, HANDLE hTXLog, int iThreadID)
{
	STATUS			stAPIRC;									//  Return code from Notes API calls
	TIMEDATE		tdCurrent;									//  Current time/date
	INTLFORMAT		ifCurrent;									//  International settings
	TFMT			tfLogTime;									//  Format for log timestamps
	WORD			wRetLen;									//  Returned length
	WORD			wLastEntryLen;								//  Length of the last entry
	WORD			wEntries;									//  Entries in the test list
	int				iIndex;										//  Generic index
	void			*pList;										//  Pointer to the list structure
	char			*pLastEntry;								//  Pointer to the last entry
	char			szTimeStamp[MAXALPHATIMEDATE + 1];			//  Formatted timestamp	
	char			szDecoratedMsg[MAX_MSG + 1];				//  Buffer for the decorated message

	if (hTXLog != NULLHANDLE)
	{

		//  Acquire the History MUTEX
		while (!xeLocal->AcquireAppMutex(HISTORY_MUTEX, iThreadID));

		//  Decorate the message with the timestamp and thread id
		OSGetIntlSettings(&ifCurrent, sizeof(ifCurrent));
		tfLogTime.Date = TDFMT_FULL;
		tfLogTime.Time = TTFMT_FULL;
		tfLogTime.Zone = TZFMT_ALWAYS;                           //  Always show Zone
		tfLogTime.Structure = TSFMT_DATETIME;
	
		OSCurrentTIMEDATE(&tdCurrent);
		stAPIRC = ConvertTIMEDATEToText(&ifCurrent,
										&tfLogTime,
										&tdCurrent,
										szTimeStamp,
										sizeof(szTimeStamp),
										&wRetLen);
		szTimeStamp[wRetLen] = 0;
		sprintf_s(szDecoratedMsg, MAX_MSG, "%s: %s [%i]", szTimeStamp, szMsg, iThreadID);

		//  Append the item to the Transaction History Log

		//  Check that we still hold the mutex
		if (xeLocal->AcquireAppMutex(HISTORY_MUTEX, iThreadID))
		{
			pList = OSLockObject(hTXLog);
			wEntries = ListGetNumEntries(pList, FALSE);
			wRetLen = ListGetSize(pList, FALSE);

			//  Get a pointer to the last entry in the list to determine if it is already truncated
			pLastEntry = szDecoratedMsg;
			wLastEntryLen = (WORD) strlen(LOG_TRUNCATED_MSG);
			if (wEntries > 0) ListGetText(pList, FALSE, wEntries - 1, &pLastEntry, &wLastEntryLen);
			if (memcmp(pLastEntry, LOG_TRUNCATED_MSG, wLastEntryLen) != 0)
			{
				//  Attempt to add the new entry
				stAPIRC = ListAddEntry(hTXLog, FALSE, &wRetLen, wEntries, szDecoratedMsg, (WORD) strlen(szDecoratedMsg));
				if (ERR(stAPIRC) != NOERROR)
				{
					//  Failed to add another message - remove the last 3 messages and mark the log truncated
					xeLocal->LogMessage(MSG_QMV0307E);
					xeLocal->GetAPIMessage(ERR(stAPIRC), szDecoratedMsg);
					xeLocal->LogMessage(szDecoratedMsg);
					if (wEntries > 3)
					{
						for (iIndex = 0; iIndex < 3; iIndex++)
						{
							OSUnlockObject(hTXLog);
							pList = OSLockObject(hTXLog);
							wEntries = ListGetNumEntries(pList, FALSE);
							wRetLen = ListGetSize(pList, FALSE);
							stAPIRC = ListRemoveEntry(hTXLog, FALSE, &wRetLen, wEntries - 1);
						}
					}
					OSUnlockObject(hTXLog);
					pList = OSLockObject(hTXLog);
					wEntries = ListGetNumEntries(pList, FALSE);
					wRetLen = ListGetSize(pList, FALSE);
					stAPIRC = ListAddEntry(hTXLog, FALSE, &wRetLen, wEntries, LOG_TRUNCATED_MSG, (WORD) strlen(LOG_TRUNCATED_MSG));
				}
			}
			OSUnlockObject(hTXLog);
			//  Free the History Mutex
			xeLocal->FreeAppMutex(HISTORY_MUTEX, iThreadID);
		}
	}

	xeLocal->LogMessage(szMsg, iThreadID);
	return;
}

//  AddHistory
//
//  This function will append the transaction history log to the transaction document.
//
//  Parameters
//
//  1  -  HANDLE				-  Handle of the Transaction History Text List
//  2  -  NOTEHANDLE			-  Handle of the Transaction Note
//  3  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//
//  API Use Notes:
//
void AppTransactionHandler::AddHistory(HANDLE hTXLog, HANDLE hnTX, int iThreadID)
{
	STATUS			stAPIRC;									//  Return code from Notes API calls
	DWORD			dwListSize;									//  Size of the Transaction History Log list
	void			*pList;										//  Pointer to the list structure
	char			szMsg[MAX_MSG + 1];							//  Generic Message Buffer

	if (hTXLog == NULLHANDLE) return;

	pList = OSLockObject(hTXLog);
	dwListSize = (DWORD) ListGetSize(pList, FALSE);
	stAPIRC = NSFItemAppend(hnTX, 0, TRANSDOC_HISTORY, (WORD) strlen(TRANSDOC_HISTORY), TYPE_TEXT_LIST, pList, dwListSize);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_QMV0308E, iThreadID);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
	}

	OSUnlockObject(hTXLog);
	OSMemFree(hTXLog);

	return;
}

//  checkAvailability
//
//  This function will test the availability of the target server, if the server is not currently available
//  then it will move the request to the Delayed Execution Queue.
//
//  Parameters
//
//  1  -  QMoveRequest *	-  The address of the QMoveRequest
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//  int  -				STATE_EXEC_REQUEST - if the target server is available
//						STATE_WAIT_FOR_WORK - if the target server is not available
//
//  API Use Notes:
//
//
int	AppTransactionHandler::checkAvailability(QMoveRequest *mrCurrent, int iThreadID)
{
	return STATE_EXEC_REQUEST;
}

//  HasWildcards
//
//  This function will check the passed string to see if it contains any wildcard characters
//
//  Parameters
//
//  1  -  char *			-  Pointer to the string to be chacked
//
//  Returns
//
//  BOOL					-  TRUE if there are wildcards in the string
//							-  FALSE if there are no wildcards present
//
//  API Use Notes:
//
//
BOOL AppTransactionHandler::HasWildcards(char *szFileSpec)
{
	char			*pAtom;

	pAtom = szFileSpec;
	while (*pAtom != '\0')
	{
		if (*pAtom == '*') return TRUE;
		if (*pAtom == '?') return TRUE;
		pAtom++;
	}
	return FALSE;
}
