//  QMFeedProcessor
//
//  This class provides the database level exit for Domino eXplorer for processing individual
//  QFeedRequest objects.
//
//  Message Identifier:      QMV
//  Message Range:           0600 - 0699
//  Trace area assignment:   None
//  Version:                 1.4.2  Build: 37
//
//  Version History:
//
//  1.4.0		04/11/2010   -  Asynch transcation queue handler added.
// 								Incorporated RunTime (3.10.1) changes.
//  1.4.2		22/10/2011	 -	VS 2011 Updates
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2010 - 2011
//

//  Application Includes
#include "QMFeedProcessor.h"

//  Default constructor
QMFeedProcessor::QMFeedProcessor(void)
{
}

//  Normal constructor
QMFeedProcessor::QMFeedProcessor(MTExecutive *xeParent, int iThreadID)
{
	//  Save the Runtime Environment
	xeLocal = xeParent;
	ParentThread = iThreadID;
}

//  Destructor
QMFeedProcessor::~QMFeedProcessor(void)
{
}

//  ReportOnThisDb
//
//  This is the entry point used by the DX subsystem to initiate actions on a per D/B basis.
//  This function will trigger creation of a new QMove transaction for the database that was located.
//
//  Parameters
//
//  1  -  DXRequest *		-  Pointer to DXRequest that is currently being serviced
//  2  -  DBHANDLE			-  Handle of the database
//  3  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  BOOL	-  TRUE if ok to continue, FALSE if an error was encountered
//
//  Notes:
//
BOOL QMFeedProcessor::ReportOnThisDb(DXRequest *dxrRQ, HANDLE hdbEntity, int iThreadID)
{
	STATUS				stAPIRC;										//  Return code from Notes API calls
	NOTEHANDLE			hnNewTX;										//  New transaction note
	QMoveRequest		*qmrNew;										//  New Q Move Request
	QFeedRequest		*qfrCurrent;									//  Current Q Feed Request
	char				szMsg[MAX_MSG + 1];								//  Generic message buffer

	//  Cast the DXrequest to its extended QFeedRequest
	qfrCurrent = (QFeedRequest *) dxrRQ;

	//  Trace building the transaction
	sprintf_s(szMsg, MAX_MSG, MSG_QMV0606T, qfrCurrent->szDirectory, qfrCurrent->szDatabase);
	xeLocal->LogTrace(TRACE_APP, szMsg, iThreadID);

	//  Create a new QMoveRequest & populate it from the original Feeder & the DX information
	qmrNew = new QMoveRequest();
	if (!PopulateRequest(qmrNew, qfrCurrent, iThreadID))
	{
		sprintf_s(szMsg, MAX_MSG, MSG_QMV0602E, dxrRQ->szDirectory, dxrRQ->szDatabase);
		LogTransactionMessage(szMsg, qfrCurrent->qmrFeeder->hTXLog, iThreadID);
		delete qmrNew;
		return TRUE;
	}

	//  Create a new note in the repository to serialize the transaction
	stAPIRC = NSFNoteCreate(xeLocal->GetMappedDBH(xeLocal->hdbRepository, iThreadID), &hnNewTX);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_QMV0603E, dxrRQ->szDirectory, dxrRQ->szDatabase);
		LogTransactionMessage(szMsg, qfrCurrent->qmrFeeder->hTXLog, iThreadID);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		LogTransactionMessage(szMsg, qfrCurrent->qmrFeeder->hTXLog, iThreadID);
		delete qmrNew;
		return TRUE;
	}

	//  Seralize the transaction data to the new request document
	if (!qmrNew->serialize(0, hnNewTX, xeLocal))
	{
		sprintf_s(szMsg, MAX_MSG, MSG_QMV0604E, dxrRQ->szDirectory, dxrRQ->szDatabase);
		LogTransactionMessage(szMsg, qfrCurrent->qmrFeeder->hTXLog, iThreadID);
		NSFNoteClose(hnNewTX);
		delete qmrNew;
		return TRUE;
	}

	//  Set the Urgent & Approved flags
	NSFItemSetText(hnNewTX, TRANSDOC_URGENT_FLAG, "0", MAXWORD);
	//  TEMP: Set not approved for testing
	NSFItemSetText(hnNewTX, TRANSDOC_APPROVED_FLAG, "Yes", MAXWORD);
	//  Set the status to new
	NSFItemSetText(hnNewTX, TRANSDOC_STATUS, TRANSDOC_STATUS_NEW, MAXWORD);

	//  Save and close the new transaction
	NSFNoteUpdate(hnNewTX, UPDATE_FORCE);
	NSFNoteClose(hnNewTX);
	
	//  Increment the count of transactions created and clean up
	sprintf_s(szMsg, MAX_MSG, MSG_QMV0605I, qmrNew->szTransID, qmrNew->szSourceDatabase);
	xeLocal->LogVerbose(szMsg, iThreadID);
	qfrCurrent->dwTransactionsCreated++;
	delete qmrNew;
	return TRUE;
}

//  PopulateRequest
//
//  This function will populate a new QMoveRequest object from a QFeed request that has been returned by the
//  Domino eXplorer engine.
//
//  Parameters
//
//  1  -  QMoveRequest *	-  Pointer to QMoveRequest that will be populated
//  2  -  QFeedRequest *	-  Pointer to the QFeedRequest
//  3  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  BOOL	-  TRUE if ok to continue, FALSE if an error was encountered
//
//  Notes:
//
BOOL QMFeedProcessor::PopulateRequest(QMoveRequest *qmrRQ, QFeedRequest *qfrCurrent, int iThreadID)
{
	char			szMsg[MAX_MSG + 1];				//  Generic Message Buffer

	//  Set the Source Database
	if (qfrCurrent->szDirectory[0] == '\0')
	{
		strcpy_s(qmrRQ->szSourceDatabase, MAX_DATABASE, qfrCurrent->szDatabase);
	}
	else
	{
		sprintf_s(qmrRQ->szSourceDatabase, MAX_DATABASE, "%s/%s", qfrCurrent->szDirectory, qfrCurrent->szDatabase);
	}
	
	//  Set the Target Directory
	strcpy_s(qmrRQ->szTargetDir, MAX_DATABASE, qfrCurrent->qmrFeeder->szTargetDir);

	//  Build and set the transaction ID from the base transaction ID plus a unique ID.
	xeLocal->GenUniqueID(szMsg, MAX_MSG, iThreadID);
	if ((strlen(szMsg) + strlen(qfrCurrent->qmrFeeder->szTransID) + 1) > MAX_DATABASE)
	{
		qfrCurrent->qmrFeeder->szTransID[MAX_DATABASE - (strlen(szMsg) + 1)] = '\0';
	}
	sprintf_s(qmrRQ->szTransID, MAX_DATABASE, "%s-%s", qfrCurrent->qmrFeeder->szTransID, szMsg);

	//  Copy all of the other settings from the Q Feed transaction into the new transaction object
	qmrRQ->RetryCount = 0;
	qmrRQ->RetryLimit = qfrCurrent->qmrFeeder->RetryLimit;
	qmrRQ->RestartActions = qfrCurrent->qmrFeeder->RestartActions;
	qmrRQ->FaultTolerance = qfrCurrent->qmrFeeder->FaultTolerance;
	qmrRQ->AllowRetry = qfrCurrent->qmrFeeder->AllowRetry;
	qmrRQ->PerformCopy = qfrCurrent->qmrFeeder->PerformCopy;
	qmrRQ->PerformDelete = qfrCurrent->qmrFeeder->PerformDelete;
	qmrRQ->PerformLink = qfrCurrent->qmrFeeder->PerformLink;

	return TRUE;
}

//  GetDescription
//
//  This is the entry point used by the DX subsystem to obtain a descriptive string for the module.
//
//  Parameters
//
//  1  -  char *		-  Pointer to the buffer where the description string will be returned
//
//  Returns
//
//  Notes:
//
void QMFeedProcessor::GetDescription(char *szDescription)
{
	//  Show the description of this Reporter
	strcpy_s(szDescription, MAX_DESC, REPORTER_DESC);
	return;
}

//  LogTransactionMessage
//
//  This function will append a decorated version of the message to the Transaction History Log.
//  It will also pass the message to the system logger.
//
//  Parameters
//
//  1  -  char *				-  Pointer to the undecorated message text
//  2  -  HANDLE				-  Handle of the Transaction History Text List
//  3  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//
//  API Use Notes:
//
void QMFeedProcessor::LogTransactionMessage(char *szMsg, HANDLE hTXLog, int iThreadID)
{
	STATUS			stAPIRC;									//  Return code from Notes API calls
	TIMEDATE		tdCurrent;									//  Current time/date
	INTLFORMAT		ifCurrent;									//  International settings
	TFMT			tfLogTime;									//  Format for log timestamps
	WORD			wRetLen;									//  Returned length
	WORD			wLastEntryLen;								//  Length of the last entry
	WORD			wEntries;									//  Entries in the test list
	int				iIndex;										//  Generic index
	void			*pList;										//  Pointer to the list structure
	char			*pLastEntry;								//  Pointer to the last entry
	char			szTimeStamp[MAXALPHATIMEDATE + 1];			//  Formatted timestamp	
	char			szDecoratedMsg[MAX_MSG + 1];				//  Buffer for the decorated message

	if (hTXLog != NULLHANDLE)
	{

		//  Acquire the History MUTEX
		while (!xeLocal->AcquireAppMutex(HISTORY_MUTEX, iThreadID));

		//  Decorate the message with the timestamp and thread id
		OSGetIntlSettings(&ifCurrent, sizeof(ifCurrent));
		tfLogTime.Date = TDFMT_FULL;
		tfLogTime.Time = TTFMT_FULL;
		tfLogTime.Zone = TZFMT_ALWAYS;                           //  Always show Zone
		tfLogTime.Structure = TSFMT_DATETIME;
	
		OSCurrentTIMEDATE(&tdCurrent);
		stAPIRC = ConvertTIMEDATEToText(&ifCurrent,
										&tfLogTime,
										&tdCurrent,
										szTimeStamp,
										sizeof(szTimeStamp),
										&wRetLen);
		szTimeStamp[wRetLen] = 0;
		sprintf_s(szDecoratedMsg, MAX_MSG, "%s: %s [%i]", szTimeStamp, szMsg, iThreadID);

		//  Append the item to the Transaction History Log

		//  Check that we still hold the mutex
		if (xeLocal->AcquireAppMutex(HISTORY_MUTEX, iThreadID))
		{
			pList = OSLockObject(hTXLog);
			wEntries = ListGetNumEntries(pList, FALSE);
			wRetLen = ListGetSize(pList, FALSE);

			//  Get a pointer to the last entry in the list to determine if it is already truncated
			pLastEntry = szDecoratedMsg;
			wLastEntryLen = (WORD) strlen(LOG_TRUNCATED_MSG);
			if (wEntries > 0) ListGetText(pList, FALSE, wEntries - 1, &pLastEntry, &wLastEntryLen);
			if (memcmp(pLastEntry, LOG_TRUNCATED_MSG, wLastEntryLen) != 0)
			{
				//  Attempt to add the new entry
				stAPIRC = ListAddEntry(hTXLog, FALSE, &wRetLen, wEntries, szDecoratedMsg, (WORD) strlen(szDecoratedMsg));
				if (ERR(stAPIRC) != NOERROR)
				{
					//  Failed to add another message - remove the last 3 messages and mark the log truncated
					xeLocal->LogMessage(MSG_QMV0601E);
					xeLocal->GetAPIMessage(ERR(stAPIRC), szDecoratedMsg);
					xeLocal->LogMessage(szDecoratedMsg);
					if (wEntries > 3)
					{
						for (iIndex = 0; iIndex < 3; iIndex++)
						{
							OSUnlockObject(hTXLog);
							pList = OSLockObject(hTXLog);
							wEntries = ListGetNumEntries(pList, FALSE);
							wRetLen = ListGetSize(pList, FALSE);
							stAPIRC = ListRemoveEntry(hTXLog, FALSE, &wRetLen, wEntries - 1);
						}
					}
					OSUnlockObject(hTXLog);
					pList = OSLockObject(hTXLog);
					wEntries = ListGetNumEntries(pList, FALSE);
					wRetLen = ListGetSize(pList, FALSE);
					stAPIRC = ListAddEntry(hTXLog, FALSE, &wRetLen, wEntries, LOG_TRUNCATED_MSG, (WORD) strlen(LOG_TRUNCATED_MSG));
				}
			}
			OSUnlockObject(hTXLog);
			//  Free the History Mutex
			xeLocal->FreeAppMutex(HISTORY_MUTEX, iThreadID);
		}
	}

	xeLocal->LogMessage(szMsg, iThreadID);
	return;
}
