//  QMFeeder
//
//  This class provides the capability of handling "Feeder" transactions within QMove.
//  Feeder transaction are those that specify a higher level of source/target than database level and
//  therefore result in scanning to generate database level transactions.
//
//  Message Identifier:      QMV
//  Message Range:           0400 - 0499
//  Trace area assignment:   None
//  Version:                 1.4.2  Build: 37
//
//  Version History:
//
//  1.4.0		17/11/2010   -  Asynch transcation queue handler added.
// 								Incorporated RunTime (3.10.1) changes.
//  1.4.2		22/10/2011	 -	VS 2011 Updates
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2010 - 2011
//

//  Application Includes
#include "QMFeeder.h"

//  Default constructor
QMFeeder::QMFeeder(void)
{
}

//  Normal constructor
QMFeeder::QMFeeder(MTExecutive *xeParent, int iThreadID)
{
	//  Save the Runtime Environment
	xeLocal = xeParent;
	ParentThread = iThreadID;
	bOnDemandInitDone = FALSE;								//  On Demand initialisation not done
	dxLocal = NULL;
	qpLocal = NULL;
}

//  Destructor
QMFeeder::~QMFeeder(void)
{
	//  If the Domino Explorer was created then delete it
	if (dxLocal != NULL) delete dxLocal;
	if (qpLocal != NULL) delete qpLocal;
}

//  ProcessThisRequest
//
//  This function will execute a single "Feeder" transaction
//
//  Parameters
//
//  1  -  QMoveRequest *	-  Pointer to the "Feeder" request
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void QMFeeder::ProcessThisRequest(QMoveRequest *qmrRQ, int iThreadID)
{
	QFeedRequest	*qfrNew;									//  Queue Feed Request
	QFeedRequest	*qfrReturned;								//  Queue Feed Request - returned from rejoin
	int				iRejoinRC;									//  Return code from Rejoin calls
	char			szMsg[MAX_MSG + 1];							//  Generic Message Buffer

	//  Firstly determine if the On-Demand initialisation has been done - if not do it.
	if (!bOnDemandInitDone) bOnDemandInitDone = DoLazyInitialisation(iThreadID);

	//  If the initialisation did not work then just fail the transaction
	if (!bOnDemandInitDone)
	{
		LogTransactionMessage(MSG_QMV0401E, qmrRQ->hTXLog, iThreadID);
		qmrRQ->ReturnCode = RETURN_INCOMPLETE;
		return;
	}

	//  Create the QFeedRequest object that will be passed to the Domino eXplorer engine 
	qfrNew = new QFeedRequest();
	qfrNew->qmrFeeder = qmrRQ;									//  Pass the address of the original QC request
	//  Set the QFeeder Exit in the request
	qfrNew->ScannersNeeded = SCANNER_DATABASE;					//  Exit is for the D/B Level
	qfrNew->Reporters = DXREP_DATABASE;							//  Reporter is for D/B Exits
	qfrNew->dxrDatabase = qpLocal;								//  Set the object address
	qfrNew->hTXLog = qmrRQ->hTXLog;								//  Pass the transaction log

	//  Initialise the DX parts of the request
	if (!qfrNew->ParseRequest(xeLocal, "Local", qmrRQ->szSourceDatabase, iThreadID))
	{
		//  Parsing of the request was not accepted by DX
		LogTransactionMessage(MSG_QMV0402E, qmrRQ->hTXLog, iThreadID);
		delete qfrNew;
		qmrRQ->ReturnCode = RETURN_INCOMPLETE;
		return;

	}

	//  Dispatch the Feed Request for the Domino Explorer to execute.
	if (!xeLocal->PostARequest(PXR_WAITIF_BUSY | PXR_APP_WAIT,
								this,
								dxLocal,
								qfrNew,
								RQATTR_REJOIN | RQATTR_CMLSL1,
								20,
								iThreadID))
	{
		//  This is an SNO (Should Not Occur) condition - this interface will block on resource shortage
		LogTransactionMessage(MSG_QMV0403E, qmrRQ->hTXLog, iThreadID);
		delete qfrNew;
		qmrRQ->ReturnCode = RETURN_INCOMPLETE;
		return;
	}

	//  Loop processing the rejoin queue until all dispatched requests have completed

	iRejoinRC = xeLocal->GetRejoinRequest(this, (void **) &qfrReturned, iThreadID);
	while (iRejoinRC != RJR_NONE_EXIST)
	{
		if (iRejoinRC == RJR_RETURNED)
		{
			//  Capture the count of transactions generated
			qfrReturned->qmrFeeder->dwDataDocsCopied = qfrReturned->dwTransactionsCreated;
			qfrReturned->qmrFeeder->ReturnCode = RETURN_NOERROR;
			sprintf_s(szMsg, MAX_MSG, MSG_QMV0405I, qfrReturned->dwTransactionsCreated);
			LogTransactionMessage(szMsg, qfrReturned->qmrFeeder->hTXLog, iThreadID);
			delete qfrReturned;
			qfrReturned = NULL;
		}
		else xeLocal->AppWait(PXR_BUSY_WAIT);
		iRejoinRC = xeLocal->GetRejoinRequest(this, (void **) &qfrReturned, iThreadID);
	}

	return;
}

//  DoLazyInitialisation
//
//  This function will perform the on-demand initialisation for the Feeder engine
//
//  Parameters
//
//  1  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  BOOL		-  TRUE if the initialisation completed
//				   FALSE if the initialisation failed
//
//  API Use Notes:
//
//
BOOL QMFeeder::DoLazyInitialisation(int iThreadID)
{
	dxLocal = new DominoExplorer(xeLocal, 1, iThreadID);
	qpLocal = new QMFeedProcessor(xeLocal, iThreadID);
	return TRUE;
}

//  ExecuteThisRequest
//
//  This is the entry point used by the multi-threading kernel to trigger processing in different threadpool threads.
//
//  Parameters
//
//  1  -  void *			-  Pointer to the parameter object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void QMFeeder::ExecuteThisRequest(void *pReqObject, int iThreadID)
{
	PartCopyRequest		*pcrCurrent;

	pcrCurrent = (PartCopyRequest *) pReqObject;
	ProcessThisRequest((QMoveRequest *) pcrCurrent->Parent, iThreadID);
	return;
}

//  LogTransactionMessage
//
//  This function will append a decorated version of the message to the Transaction History Log.
//  It will also pass the message to the system logger.
//
//  Parameters
//
//  1  -  char *				-  Pointer to the undecorated message text
//  2  -  HANDLE				-  Handle of the Transaction History Text List
//  3  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//
//  API Use Notes:
//
void QMFeeder::LogTransactionMessage(char *szMsg, HANDLE hTXLog, int iThreadID)
{
	STATUS			stAPIRC;									//  Return code from Notes API calls
	TIMEDATE		tdCurrent;									//  Current time/date
	INTLFORMAT		ifCurrent;									//  International settings
	TFMT			tfLogTime;									//  Format for log timestamps
	WORD			wRetLen;									//  Returned length
	WORD			wLastEntryLen;								//  Length of the last entry
	WORD			wEntries;									//  Entries in the test list
	int				iIndex;										//  Generic index
	void			*pList;										//  Pointer to the list structure
	char			*pLastEntry;								//  Pointer to the last entry
	char			szTimeStamp[MAXALPHATIMEDATE + 1];			//  Formatted timestamp	
	char			szDecoratedMsg[MAX_MSG + 1];				//  Buffer for the decorated message

	if (hTXLog != NULLHANDLE)
	{

		//  Acquire the History MUTEX
		while (!xeLocal->AcquireAppMutex(HISTORY_MUTEX, iThreadID));

		//  Decorate the message with the timestamp and thread id
		OSGetIntlSettings(&ifCurrent, sizeof(ifCurrent));
		tfLogTime.Date = TDFMT_FULL;
		tfLogTime.Time = TTFMT_FULL;
		tfLogTime.Zone = TZFMT_ALWAYS;                           //  Always show Zone
		tfLogTime.Structure = TSFMT_DATETIME;
	
		OSCurrentTIMEDATE(&tdCurrent);
		stAPIRC = ConvertTIMEDATEToText(&ifCurrent,
										&tfLogTime,
										&tdCurrent,
										szTimeStamp,
										sizeof(szTimeStamp),
										&wRetLen);
		szTimeStamp[wRetLen] = 0;
		sprintf_s(szDecoratedMsg, MAX_MSG, "%s: %s [%i]", szTimeStamp, szMsg, iThreadID);

		//  Append the item to the Transaction History Log

		//  Check that we still hold the mutex
		if (xeLocal->AcquireAppMutex(HISTORY_MUTEX, iThreadID))
		{
			pList = OSLockObject(hTXLog);
			wEntries = ListGetNumEntries(pList, FALSE);
			wRetLen = ListGetSize(pList, FALSE);

			//  Get a pointer to the last entry in the list to determine if it is already truncated
			pLastEntry = szDecoratedMsg;
			wLastEntryLen = (WORD) strlen(LOG_TRUNCATED_MSG);
			if (wEntries > 0) ListGetText(pList, FALSE, wEntries - 1, &pLastEntry, &wLastEntryLen);
			if (memcmp(pLastEntry, LOG_TRUNCATED_MSG, wLastEntryLen) != 0)
			{
				//  Attempt to add the new entry
				stAPIRC = ListAddEntry(hTXLog, FALSE, &wRetLen, wEntries, szDecoratedMsg, (WORD) strlen(szDecoratedMsg));
				if (ERR(stAPIRC) != NOERROR)
				{
					//  Failed to add another message - remove the last 3 messages and mark the log truncated
					xeLocal->LogMessage(MSG_QMV0404E);
					xeLocal->GetAPIMessage(ERR(stAPIRC), szDecoratedMsg);
					xeLocal->LogMessage(szDecoratedMsg);
					if (wEntries > 3)
					{
						for (iIndex = 0; iIndex < 3; iIndex++)
						{
							OSUnlockObject(hTXLog);
							pList = OSLockObject(hTXLog);
							wEntries = ListGetNumEntries(pList, FALSE);
							wRetLen = ListGetSize(pList, FALSE);
							stAPIRC = ListRemoveEntry(hTXLog, FALSE, &wRetLen, wEntries - 1);
						}
					}
					OSUnlockObject(hTXLog);
					pList = OSLockObject(hTXLog);
					wEntries = ListGetNumEntries(pList, FALSE);
					wRetLen = ListGetSize(pList, FALSE);
					stAPIRC = ListAddEntry(hTXLog, FALSE, &wRetLen, wEntries, LOG_TRUNCATED_MSG, (WORD) strlen(LOG_TRUNCATED_MSG));
				}
			}
			OSUnlockObject(hTXLog);
			//  Free the History Mutex
			xeLocal->FreeAppMutex(HISTORY_MUTEX, iThreadID);
		}
	}

	xeLocal->LogMessage(szMsg, iThreadID);
	return;
}
