//     Q(ueue)Move - Queued database mover

///******************************************************************************************************************
//*																													*
//*   File:       QMove.cpp																							*
//*   Suite:      Domino Admin Utilities																			*
//*   Version:    1.4.2	build: 39																					*
//*   Author:     Ian Tree/HMNL																						*
//*																													*
//*   Copyright 2009 - 2011 Hadleigh Marshall Netherlands b.v.														*
//*******************************************************************************************************************
//*	Q(ueue)Move																										*
//*																													*
//*	This utility server add-in task will move databases safely around storage while the databases are in use.		*
//*																													*
//*	USAGE:																											*
//*																													*
//*	Load QMove ControlDb [-I:nn][-B:nn][-V][-T[:nn]][-D][-E]														*
//*																													*
//*	[ControlDatabase]	- The name of the control database, relative to the server's Data Directory					*
//* [-I:nnn] - Delay between transactions 1-600 seconds.															*
//* [-B:nnnnn] - Transfer buffer size in Kb (4-1024).																*
//* [-A:nnnnn] - Copy Arena size in Mb.																				*
//* [-L] - Enable use of the large file copier.																		*
//* [-V]|[-T[:Area]]|[-D] - Set logging level to Verbose, Trace (optionall the area to trace) or Debug.				*
//*	-E					- Optional - Switch that echoes the logging messages to the server console					*
//*	=IniFilename		- Optional - Notes ini filename to use														*
//*																													*
//*	NOTES:																											*
//*																													*
//*																													*
//*******************************************************************************************************************
//*																													*
//*   History:																										*
//*																													*
//*	1.0.0 - 11/2009   -  Base Release																				*
//* 1.1.0 - 12/2009   -	 Enhanced move operation modes supported													*
//* 1.2.0 - 01/2010   -  Variable delay between transactions														*
//*					  -  Variable buffer size for copying															*
//* 1.3.0 - 02/2010   -  Handle D/B and Dir links																	*
//* 1.3.1 - 02/2010   -  Improve runtime messages																	*
//* 1.3.3 - 10/2010   -  Switch to Runtime 3.10.1																	*
//* 1.4.0 - 10/2010   -  Use Command & Transaction Handlers															*
//* 1.4.1 - 12/2010	  -  Switch to Runtime 3.10.3																	*
//*						 Fixed an issue with the transaction handler												*
//* 1.4.2 - 10/2011	  -  VS 2011 Updates																			*
//*																													*
//******************************************************************************************************************/

//  Application Includes
#include "QMove.h"

//   Main Function 

int main(int argc, char *argv[])
{
	AppRunSettings			*arsLocal;								//  Run settings for this execution
	AppCommandHandler		*achLocal;								//  Application specific MQ command handler
	AppTransactionHandler	*thLocal = NULL;						//  Application specific Transaction Queue handler
	ThreadManagerPolicy		*tmpInit;								//  Initial Thread Manager Policy
	MTExecutive				*xeLocal;								//  Runtime for this application
#if (defined(_DEBUG) && defined(WIN32))
	Helper					*DebugHelper;							//  Debugging Helper
#endif
	QMover					*qmLocal;								//  Extended D/B Mover
	QMFeeder				*qfLocal;								//  Feeder transaction processor
	TransactionQueue		*tqQMove = NULL;						//  Transaction Queue
	TransactionQueue		*tqReturned;							//  Rejoined queue entity
	BOOL					bProcessorShouldQuit;					//  Signal flag for the task to quit
	int						iRejoinRC;								//  Rejoin return code
	char					szBuildID[MAX_BUILD + 1];				//  Build Info
	char					szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Show application is starting
	std::cout << APP_TITLE << " (" << APP_NAME << ") Version: " << APP_VERSION << " is starting." << std::endl;

	//  Validate and build the Run Settings for this execution
	arsLocal = new AppRunSettings(argc, argv);
	if (!arsLocal->AllowExecution)								//  Exit silently if just showing usage
	{
		delete arsLocal;
		return APPRC_NOERROR;
	}

	if (!arsLocal->IsValid)										//  Abort if validation failed
	{
		std::cout << MSG_QMV0101S << std::endl;
		delete arsLocal;
		return APPRC_FATAL;
	}

	//  Set the identification in the Run Settings
	strcpy_s(arsLocal->APPName, MAXAPPNAME, APP_NAME);
	strcpy_s(arsLocal->APPTitle, MAXAPPTITLE, APP_TITLE);
	strcpy_s(arsLocal->APPVer, MAXAPPVERSION, APP_VERSION);

	//  Create and initialise a Thread Manager Policy
	tmpInit = new ThreadManagerPolicy();
	tmpInit->TPoolPolicy = TPOOL_POLICY_OBEYMAXMIN | TPOOL_POLICY_PRESTARTTARGET;
	tmpInit->MaxThreads = arsLocal->ThreadCount;								//  Max Thread Count
	tmpInit->MinThreads = arsLocal->ThreadCount;								//  Min Thread Count
	tmpInit->TargetThreads = arsLocal->ThreadCount;								//  Initial target Thread Count
	tmpInit->PendingRQECapacity = PCAP_PERTHREAD * arsLocal->ThreadCount;		//  Pending RQE Pool size
	tmpInit->RejoinRQECapacity = RCAP_PERTHREAD * arsLocal->ThreadCount;		//  Rejoin RQE Pool size
	tmpInit->AsyLogPoolEntries = 200;											//  Log pool size is 200 entries
	
	//  Set the priority policy
	tmpInit->PriorityPolicy = PRIO_POLICY_AGERQS | PRIO_POLICY_PREFBOOST;

	//  Set CMLS Mode and the CMLS contraints
	tmpInit->TPSchedMode = TPOOL_MODE_CMLS;
	tmpInit->MaxPctL1Threads = 20;									//  20% allocated to Feeders

	//  Initialise the runtime environment
	xeLocal = new MTExecutive(arsLocal, tmpInit, argc, argv);
	if (!xeLocal->IsInitialised)									//  Abort if runtime failed to initialise
	{
		std::cout << MSG_QMV0101S << std::endl;
		delete xeLocal;
		delete tmpInit;
		delete arsLocal;
		return APPRC_FATAL;
	}

	//  Also check the the Thread Manager was initialised properly
	if (!xeLocal->MTFunctionsAvailable)
	{
		std::cout << MSG_QMV0101S << std::endl;
		delete xeLocal;
		delete tmpInit;
		delete arsLocal;
		return APPRC_FATAL;
	}

#if (defined(_DEBUG) && defined(WIN32))
	//  Construct the Debugging Helper object
	DebugHelper = new Helper(xeLocal, 0);
	DebugHelper->ReportMemoryUsage(FALSE, FALSE, 0);
#endif

	//  Inform if the Large File Copier is disabled
	if (!arsLocal->UseLFC) xeLocal->LogMessage(MSG_QMV0109I, 0);
	else
	{
		//  Show the copy size settings
		sprintf_s(szMsg, MAX_MSG, MSG_QMV0110I, arsLocal->ArenaSize, arsLocal->TransferBufferK);
		xeLocal->LogMessage(szMsg, 0);
	}

	//  Construct the QMover move engine
	qmLocal = new QMover(xeLocal, 0);
	//  Construct the Feeder Transaction processor
	qfLocal = new QMFeeder(xeLocal, 0);

	//  Construct the Application Transaction Handler
	thLocal = new AppTransactionHandler(xeLocal, 0);
	thLocal->qmMaster = qmLocal;							//  Set the move engine in the transaction handler
	thLocal->qfMaster = qfLocal;							//  Set the feeder engine in the transaction handler

	//  Now construct a new Transaction Queue descriptor for the transaction queue
	tqQMove = new TransactionQueue();
	//  Set the D/B handle of the repository as the queue database
	tqQMove->hdbQueue = xeLocal->hdbRepository;
	//  Set the Queue Name
	strcpy_s(tqQMove->szQName, MAX_ELEMENT, QMQNAME);
	//  Define the views to use for the sub-queues
	strcpy_s(tqQMove->szReadyQName, MAX_ELEMENT, QUEUE_NEW_TRANS_VIEW);
	strcpy_s(tqQMove->szInProgressQName, MAX_ELEMENT, QUEUE_IP_TRANS_VIEW);
	strcpy_s(tqQMove->szDelayedQName, MAX_ELEMENT, QUEUE_DELAYED_TRANS_VIEW);
	//  Set the options that control the queue behaviours
	tqQMove->wQueueProtocols = QPFLAG_DEFAULT;				//  Default Behaviour for startup & delay
	tqQMove->MaxConcurrent = arsLocal->MaxCTX;				//  Set initial Max Concurrency
	tqQMove->MinDelay = 15 * 60;							//  Delay Queue for minimum of 15 mins
	tqQMove->DelayCycleSecs = 10;							//  10 seconds per dispatch loop

	//  Contruct the Application Command Handler and attach it to the Thread Monitor
	achLocal = new AppCommandHandler(xeLocal);
	achLocal->tqCurrent = tqQMove;							//  Set the Queue address in the handler
	achLocal->thCurrent = thLocal;							//  Set the Queue handler address
	xeLocal->AttachCommandHandler(achLocal);

#if (defined(_DEBUG) && defined(WIN32))
	//  Activate the debug command extensions
	achLocal->DebugHelper = DebugHelper;
	thLocal->DebugHelper = DebugHelper;
#endif

	//  Report the engine build
	xeLocal->GetBuildInfo_s(szBuildID, MAX_BUILD);
	sprintf_s(szMsg, MAX_MSG, MSG_QMV0102I, szBuildID);
	xeLocal->LogVerbose(szMsg, 0);

	//  Dispatch the Transaction Queue for execution
	xeLocal->PostARequest(PXR_WAITIF_BUSY | PXR_APP_WAIT,
							achLocal,
							thLocal,
							tqQMove,
							RQATTR_REJOIN | RQATTR_CMLSL0,
							10,
							0);

	//  Main Processing (or not) Loop
	//  Monitor the command handler and respond to any requested changes in state
	bProcessorShouldQuit = FALSE;
	AddInSetStatusText(ADDIN_STATUS_RUNNING);
	while (!bProcessorShouldQuit)
	{
		//  Check if any terminal condition signals have been posted in the Command Handler
		if (achLocal->State & CH_STATE_QUIT)
		{
			//  Detach the command handler 
			xeLocal->AttachCommandHandler(NULL);
			AddInSetStatusText(ADDIN_STATUS_TERM);
			xeLocal->LogMessage(MSG_QMV0107I); 
			AddInLogMessageText(MSG_QMV0107I, NOERROR);
			bProcessorShouldQuit = TRUE;
		}
		else
		{
			if (achLocal->State & (CH_STATE_STOP | CH_STATE_ABORT))
			{
				//  Detach the command handler 
				xeLocal->AttachCommandHandler(NULL);
				AddInSetStatusText(ADDIN_STATUS_TERM);
				xeLocal->LogMessage(MSG_QMV0108I); 
				AddInLogMessageText(MSG_QMV0108I, NOERROR);
				bProcessorShouldQuit = TRUE;
			}
			else xeLocal->AppWait(TIMEOUT_EVENT_WAIT);				//  Sleep
		}
	}

	//  Wait for the transaction queue to drain before shutting down
	tqQMove->LocalPermit = FALSE;
	iRejoinRC = xeLocal->GetRejoinRequest(achLocal, (void **) &tqReturned, 0);
	while (iRejoinRC != RJR_NONE_EXIST)
	{	
		if (iRejoinRC != RJR_RETURNED) xeLocal->AppWait(PXR_BUSY_WAIT);
		iRejoinRC = xeLocal->GetRejoinRequest(achLocal, (void **) &tqReturned, 0);
	}

	sprintf_s(szMsg, MAX_MSG, MSG_QMV0106I, tqQMove->szQName);
	xeLocal->LogMessage(szMsg, 0);

	//  Dispose of the inactive components
	if (achLocal != NULL) delete achLocal;
	if (tqQMove != NULL) delete tqQMove;
	if (thLocal != NULL) delete thLocal;
	if (qmLocal != NULL) delete qmLocal;
	if (qfLocal != NULL) delete qfLocal;

#if (defined(_DEBUG) && defined(WIN32))
	DebugHelper->ReportMemoryUsage(FALSE, FALSE, 0);
	delete DebugHelper;
#endif

	//  Processing is completed - shut down, clean up and exit.
	xeLocal->ShowStats(STATS_DEBUG | STATS_THREAD, 0);
	if (!xeLocal->Close())
	{
		std::cout << MSG_QMV0103S << std::endl;
		//  Cleanup the local objects
		delete xeLocal;
		delete tmpInit;
		delete arsLocal;
#if (defined(_DEBUG) && defined(WIN32))
		CheckMemoryLeaks();
#endif
		return APPRC_FATAL;
	}

	//  Cleanup the local objects
	delete xeLocal;
	delete tmpInit;
	delete arsLocal;

	//  Terminate the application
	std::cout << APP_TITLE << " (" << APP_NAME << ") Version: " << APP_VERSION << " has completed normally." << std::endl;
#if (defined(_DEBUG) && defined(WIN32))
	CheckMemoryLeaks();
#endif
	return APPRC_NOERROR;
}

