//  QMover
//
//  This class extends the DX DbMover class to provide additional Database Move services that
//  are needed by the QMove application.
//
//  Version:                 1.4.2  Build: 37
//
//  Version History:
//
//  1.0.0		- 14/11/2009			Initial Version
//  1.1.0		- 01/12/2009		-	Enhanced move operation modes supported
//  1.2.0		- 01/01/2010		-	Variable delay between transactions
//									-	Variable buffer size for copying
//  1.4.0		- 15/11/2010		-   MTIO support added
//  1.4.2		- 22/10/2011		-	VS 2011 Updates
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2009 - 2011
//

//  Application Includes
#include "QMover.h"

QMover::QMover(MTExecutive *xeParent, int iThreadID)
{
	xeLocal = xeParent;								//  Set the runtime
	OwningThread = iThreadID;						//  Set the owning thread
}

QMover::~QMover(void)
{
}

//  ProxyLogMessage
//
//  This implementation of the log interface will write the message to the request history field as well
//  as writing it to the standard runtime logger.
//
//  Parameters
//
//  1  -  MoveRequest *		-  Pointer to the move request
//  1  -  char *			-  Pointer to the message to be logged
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//  BOOL					-  TRUE if the log action succeeded
//							-  FALSE if the log action failed
//
//  API Use Notes:
//
//
BOOL QMover::ProxyLogMessage(MoveRequest *mrRQ, char *szMsg, int iThreadID)
{
	QMoveRequest	*qmrRQ;										//  Queue Move Request
	STATUS			stAPIRC;									//  Return code from Notes API calls
	HANDLE			hTXLog;										//  Handle to the transaction log
	TIMEDATE		tdCurrent;									//  Current time/date
	INTLFORMAT		ifCurrent;									//  International settings
	TFMT			tfLogTime;									//  Format for log timestamps
	WORD			wRetLen;									//  Returned length
	WORD			wLastEntryLen;								//  Length of the last entry
	WORD			wEntries;									//  Entries in the test list
	int				iIndex;										//  Generic index
	void			*pList;										//  Pointer to the list structure
	char			*pLastEntry;								//  Pointer to the last entry
	char			szTimeStamp[MAXALPHATIMEDATE + 1];			//  Formatted timestamp	
	char			szDecoratedMsg[MAX_MSG + 1];				//  Buffer for the decorated message

	//  Cast the passed object to the Queue Move Request
	qmrRQ = (QMoveRequest *) mrRQ;
	hTXLog = qmrRQ->hTXLog;

	if (hTXLog != NULLHANDLE)
	{
		//  Acquire the History MUTEX
		while (!xeLocal->AcquireAppMutex(HISTORY_MUTEX, iThreadID));

		//  Decorate the message with the timestamp and thread id
		OSGetIntlSettings(&ifCurrent, sizeof(ifCurrent));
		tfLogTime.Date = TDFMT_FULL;
		tfLogTime.Time = TTFMT_FULL;
		tfLogTime.Zone = TZFMT_ALWAYS;                           //  Always show Zone
		tfLogTime.Structure = TSFMT_DATETIME;
	
		OSCurrentTIMEDATE(&tdCurrent);
		stAPIRC = ConvertTIMEDATEToText(&ifCurrent,
										&tfLogTime,
										&tdCurrent,
										szTimeStamp,
										sizeof(szTimeStamp),
										&wRetLen);
		szTimeStamp[wRetLen] = 0;
		sprintf_s(szDecoratedMsg, MAX_MSG, "%s: %s [%i]", szTimeStamp, szMsg, iThreadID);

		//  Append the item to the Transaction History Log

		//  Check that we still hold the mutex
		if (xeLocal->AcquireAppMutex(HISTORY_MUTEX, iThreadID))
		{
			pList = OSLockObject(hTXLog);
			wEntries = ListGetNumEntries(pList, FALSE);
			wRetLen = ListGetSize(pList, FALSE);

			//  Get a pointer to the last entry in the list to determine if it is already truncated
			pLastEntry = szDecoratedMsg;
			wLastEntryLen = (WORD) strlen(LOG_TRUNCATED_MSG);
			if (wEntries > 0) ListGetText(pList, FALSE, wEntries - 1, &pLastEntry, &wLastEntryLen);
			if (memcmp(pLastEntry, LOG_TRUNCATED_MSG, wLastEntryLen) != 0)
			{
				//  Attempt to add the new entry
				stAPIRC = ListAddEntry(hTXLog, FALSE, &wRetLen, wEntries, szDecoratedMsg, (WORD) strlen(szDecoratedMsg));
				if (ERR(stAPIRC) != NOERROR)
				{
					//  Failed to add another message - remove the last 3 messages and mark the log truncated
					xeLocal->LogMessage(MSG_QMV0503E);
					xeLocal->GetAPIMessage(ERR(stAPIRC), szDecoratedMsg);
					xeLocal->LogMessage(szDecoratedMsg);
					if (wEntries > 3)
					{
						for (iIndex = 0; iIndex < 3; iIndex++)
						{
							OSUnlockObject(hTXLog);
							pList = OSLockObject(hTXLog);
							wEntries = ListGetNumEntries(pList, FALSE);
							wRetLen = ListGetSize(pList, FALSE);
							stAPIRC = ListRemoveEntry(hTXLog, FALSE, &wRetLen, wEntries - 1);
						}
					}
					OSUnlockObject(hTXLog);
					pList = OSLockObject(hTXLog);
					wEntries = ListGetNumEntries(pList, FALSE);
					wRetLen = ListGetSize(pList, FALSE);
					stAPIRC = ListAddEntry(hTXLog, FALSE, &wRetLen, wEntries, LOG_TRUNCATED_MSG, (WORD) strlen(LOG_TRUNCATED_MSG));
				}
			}
			OSUnlockObject(hTXLog);

			//  Free the History Mutex
			xeLocal->FreeAppMutex(HISTORY_MUTEX, iThreadID);
		}
	}

	xeLocal->LogMessage(szMsg, iThreadID);
	return TRUE;
}
